/*  NDX.C  . . .  BTree Index Handling Routines

    Copyright (C) 1987, 1988, 1989 by George H. Mealy

*/    

#ifdef TEST

/*  NDX.H   . . .   Header file for index package   */

#ifndef	NDX_H
#define NDX_H

#define STACKDEPTH  10      /* Maximum tree height */
#define NODESIZE    1024    /* Bytes per node */
#define SEEKADR     unsigned long
#define MAXKEY      128     /* Maximum key length */
#define NODUPS      0       /* No duplicate keys */
#define FIFO        1       /* ISAM, first-in-first-out */
#define LIFO        2       /* ISAM, last-in-first-out */

typedef struct{
    SEEKADR     ptr;        /* Pointer to lower level node (lesser keys) */
    SEEKADR     seekadr;     /* Record number or offset in data file */
    char        key[MAXKEY];   /* The ASCIZ key */
} KEY;

/*  Note that each node can have a final pointer to the next lower level which
    is not counted in end_keys, but is at that relative position.
*/

typedef struct{        /* Index file header */
    SEEKADR    root;       /* File offset of root node */
	SEEKADR	   next;	   /* Offset of end of file */
	SEEKADR    freelist;   /* Chain of unused nodes */
	unsigned   char minor, major; /* Major and minor version */
    unsigned   mode;       /* Retrieval mode (FIFO, etc.) */
    unsigned   nodesize;
} HDR;

typedef struct {        /* Stack frame */
    SEEKADR  node;          /* Offset of node in index file */
    unsigned off;           /* Offset of key in node->keys */
} STACK;

/* The stack holds the current retrieval state of the index:

*/

typedef struct{         /* In memory index structure */
	int (* cmpfn)();		/* Comparison function */
	KEY			key;		/* Current key */
	HDR			hdr;		/* Index file header */
	int         level;      /* Stack pointer */
	STACK		stack[STACKDEPTH];
	int         handle;     /* File handle */
    char        filename[80];   /* Filename of index (not data) file */
} INDEX;

/*	Function prototypes */

INDEX *xcreate(char * indexname, int (*compfn)(), unsigned mode);
INDEX *xopen  (char * indexname, int (*compfn)());
void xclose(INDEX * index);
SEEKADR xinsert(INDEX * index, char * key, SEEKADR seekadr);
xdelete(INDEX * index, char * key, SEEKADR seekadr);
SEEKADR xfind(INDEX * index, char * key);
SEEKADR xfirst(INDEX * index, unsigned last);
SEEKADR xnext(INDEX * index);
SEEKADR xprev(INDEX * index);

#endif	/* #ifndef NDX_H */

/*  End of NDX.H    */

#define STATIC
#else
#include    "ndx.h"
#define STATIC static       /* Hides internal code and data names */
#endif	/* #ifdef TEST */

#include	"stdio.h"
#include	"stdlib.h"
#include	"stdarg.h"
#include    "string.h"
#include	"fcntl.h"
#include	"sys\types.h"
#include    "sys\stat.h"

#ifndef 	__TURBOC__		/* MSC vs. TC and ANSI */
#include    "direct.h"
#include	"malloc.h"
#else
#include    "dir.h"
#include	"alloc.h"
#endif

#define MAJOR       2
#define MINOR       0
#define STACKTOP    ndx->level
#define HDRSIZE     512
#define NCACHE      10      /* Number of node buffers */
#define KEYLEN(k)   (2*sizeof(SEEKADR)+1+strlen(k->key))
#define NEXTKEY(k)  ((KEY *)((char *)k + KEYLEN(k)))
#define MAXDATA     (NODESIZE - 2*sizeof(SEEKADR) - sizeof(unsigned))
#define EXTRA(n)	((KEY *)(n->keys.keys + n->end_keys))
#define KEYCPY(t, s)\
    (memcpy(t, s, strlen((s)->key) + 2*sizeof(SEEKADR) + 1))
#define POP(a)              _pop(&a)
#define PUSH(a, b)          _push(a, b)

typedef struct{
    SEEKADR     seekadr;    /* Node offset or free node chain */
    unsigned    end_keys;   /* First unused data byte in keys */
	union {
		char		keys[NODESIZE-sizeof(SEEKADR)-sizeof(unsigned)];
		SEEKADR 	ptr;
		}		keys;		/* KEY keys[] */
} NODE;

typedef struct{             /* Cache entry */
    char        dirty;      /* Node needs writing */
	int         handle;     /* File handle */
	NODE        *node;      /* File block content */
} CACHE;

STATIC _xinsert(KEY * key, SEEKADR root);
STATIC SEEKADR xfindkeyrec(char *key, SEEKADR seekadr);
STATIC SEEKADR _find(char *key, SEEKADR root);
STATIC NODE *  newnode(void);
STATIC NODE *  getnode(SEEKADR n);
STATIC KEY * prevkey(NODE *node,  KEY *next);
STATIC split(NODE * node);
STATIC void shiftup(NODE * node, KEY * key, unsigned n);
STATIC void shiftdn(NODE * node, KEY * key, unsigned n, unsigned ptrup);
STATIC xwrite(SEEKADR start, void *buffer, unsigned n);
STATIC xread(SEEKADR start, void *buffer, unsigned n);
STATIC initcache();
STATIC void fatal(char *format, ...);
STATIC void _push(NODE *node, KEY *off);
STATIC NODE * _pop(KEY **off);
STATIC void xprint(INDEX *index, char *filename);

STATIC char notice[] = "B-tree indexing routines copyright 1987,88,89 by"
					"George H. Mealy.";
STATIC INDEX   *ndx;               /* Current index block */
STATIC CACHE   cache[NCACHE];      /* The node cache */
STATIC char    cache_exists;       /* Initiated if nonzero */
STATIC char    searcharg[MAXKEY];
STATIC int     searchn;

INDEX *xcreate(char * indexname, int (*compfn)(), unsigned mode)
{
	ndx = calloc(sizeof(INDEX), 1);
    ndx->handle = open(indexname, O_RDWR|O_CREAT|O_TRUNC|O_BINARY,S_IWRITE);
	if (ndx->handle == -1 || !initcache()) return 0;
    strcpy(ndx->filename, indexname);
    ndx->hdr.root = ndx->hdr.next = HDRSIZE;
    ndx->hdr.freelist = 0;
    ndx->hdr.major = MAJOR;
    ndx->hdr.minor = MINOR;
	ndx->hdr.mode = mode;
    ndx->hdr.nodesize = NODESIZE;
    newnode();
    xwrite(0L, &ndx->hdr, sizeof(HDR));
	xwrite((long)sizeof(HDR), cache[0].node->keys.keys, HDRSIZE - sizeof(HDR));
    ndx->cmpfn = compfn? compfn: strcmp;
	return ndx;
}

INDEX *xopen(char * indexname, int (*compfn)())
{
	ndx = calloc(sizeof(INDEX), 1);
    ndx->handle = open(indexname, O_RDWR|O_BINARY);
	if (ndx->handle == -1 || ! initcache()) return 0;
    strcpy(ndx->filename, indexname);
    xread(0L, &ndx->hdr, sizeof(HDR));
    ndx->cmpfn = compfn? compfn: strcmp;
	return ndx;
}

void xclose(INDEX *index)
{
	int i;

	ndx = index;
	for (i=0; i < NCACHE; i++)
        if (cache[i].handle == ndx->handle) {
            if (cache[i].dirty) 
				xwrite(cache[i].node->seekadr, cache[i].node, NODESIZE);
            cache[i].node->seekadr = cache[i].handle = cache[i].dirty = 0;
            }
    xwrite(0L, &ndx->hdr, sizeof(HDR)); /* Write header information */
    close(ndx->handle);
}

SEEKADR xinsert(INDEX * index, char * key, SEEKADR seekadr)
{
	KEY k;
    int result;
	
	ndx = index;
    if (strlen(key) >= MAXKEY) fatal("Key \"%s\" too long");
	strcpy(k.key, key);
	k.seekadr = seekadr;
	k.ptr = 0;
    while (STACKTOP = 0, result = _xinsert(&k, ndx->hdr.root) < 0);
	ndx->key = k;
    return result? k.seekadr: result;
}

STATIC int _xinsert(KEY * key, SEEKADR root)
{
	NODE * node;
    KEY *k, *ik;
	void *end;
	int compare = -1, n;
	
	if (! root) return 0;
	node = getnode(root);
	k = (KEY *)node->keys.keys;
    end = (char *)k + node->end_keys;
    while (k < end &&
      (compare = strcmp(k->key, key->key)) <= 0)
        k = NEXTKEY(k);
    if (! compare) {      /* Found it */
        ik = k;
        if (ndx->hdr.mode == NODUPS)
            {key->seekadr = k->seekadr; return 0; }
        for (; k < end; k = NEXTKEY(k)) {
				if (compare = strcmp((char *)k->key, key->key)) break;
                if (k->seekadr == key->seekadr) return k->seekadr;
			}
        if (ndx->hdr.mode == LIFO) k = ik;
        }
    if (k->ptr) {           /* Recurse */
        PUSH(node, k);
        return _xinsert(key, k->ptr);
        }
    if (node->end_keys + KEYLEN(key) > MAXDATA) /* Node full, so split */
        return split(node)? -1: 0;
	shiftup(node, k, KEYLEN(key));
    PUSH(node, NEXTKEY(k));
    KEYCPY(k, key);
	cache[0].dirty = 1;
    return k->seekadr;
}

xdelete(INDEX *index, char *key, SEEKADR seekadr)
{
    unsigned ttop, i;
    NODE *node;
    KEY *kp;
	
	ndx = index;
    if (key && !xfindkeyrec(key, seekadr)) return 0;
    ttop = STACKTOP;
    node = POP(kp);
    if (! kp->ptr) shiftdn(node, kp, KEYLEN(kp), 0);
      else {
        KEY *tkey;

        STACKTOP++;
        do  {   /* Find lower rightmost key to pull up */
            node = getnode(kp->ptr);
			kp = (KEY *)(node->keys.keys + node->end_keys);
            PUSH(node, kp);
        } while (kp->ptr);
        node = POP(kp);
        kp = prevkey(node, kp);  /* This is the key to go up */

        /* need only the seekadr and key */
        tkey = malloc(i = KEYLEN(kp));
        memcpy(tkey, kp, i);
        shiftdn(node, kp, i, 1);   /* Leaves the ptr part untouched */
        STACKTOP = ttop;
        node = POP(kp);

        /* Preserve the original ptr */
        tkey->ptr = kp->ptr;

        /* Substitute tkey for key being deleted */
        node->end_keys++;           /* Avoid deletion of node! */
        shiftdn(node, kp, KEYLEN(kp), 0);
        node->end_keys--;
        shiftup(node, kp, i);
        memcpy(kp, tkey, i);
        free(tkey);
	}
    ++STACKTOP;
    while (kp->ptr) {
        PUSH(node, kp);
        node = getnode(kp->ptr);
		kp = (KEY *)node->keys.keys;
        }
    KEYCPY(&ndx->key, kp);
    return 1;
}

STATIC SEEKADR xfindkeyrec(char *key, SEEKADR seekadr)
{
    SEEKADR rcd;

    rcd = xfind(ndx, key);
    while (rcd && rcd != seekadr) rcd = xnext(ndx);
    return rcd;
}

SEEKADR xfind(INDEX * index, char *key)
{
	ndx = index;
    if (key) {
        strcpy(searcharg, key);
        searchn = strlen(key);
        STACKTOP = 0;
        return _find(key, ndx->hdr.root);  /* Inner routine */
    }
    if (! xnext(ndx) ||
        (*ndx->cmpfn)(searcharg, ndx->key, searchn)) return NULL;
    return ndx->key.seekadr;
}

STATIC SEEKADR _find(char * key, SEEKADR root)
{
	NODE *node = getnode(root);
	KEY * k = (KEY *)node->keys.keys;
	SEEKADR rcd;
    int  compare, n = strlen(key);
	
	for (; k < (KEY *)(node->keys.keys + node->end_keys); k = NEXTKEY(k)) {
		/* Find right place in this level */
        if (! (compare = (* ndx->cmpfn)(key, k->key, n))) {
			if (k->ptr) {	/* Lower level keys -- recurse */
                PUSH(node, k);
                rcd = _find(key, k->ptr);
				if (rcd) return rcd;
                node = POP(k);
			}
				/* We have the key */
            KEYCPY(&ndx->key, k);
            PUSH(node, k);
			return ndx->key.seekadr;
		}
        if (compare < 0) break;
	}

    /* Ran into larger key or at end of this level */
    PUSH(node, k);      /* Update stack */
	rcd = 0;
	if (k->ptr) /* More keys on right -- recurse */
        rcd =  _find(key, k->ptr);
    if (! rcd) POP(k);
	return rcd;
}

SEEKADR xfirst(INDEX *index, unsigned last)
{
	NODE *node = getnode(index->hdr.root);
    SEEKADR ptr;
    KEY *kp;
	
	ndx = index;
	if (! node->end_keys && ! node->keys.ptr) return 0;
    STACKTOP = 0;
    /* Find lowermost leftmost key (or last) in the index, setting the stack */
    while (ptr =
		(kp = (KEY *)(node->keys.keys + (last? node->end_keys: 0)))->ptr) {
        PUSH(node, kp);
		node = getnode(ptr);
	}
    if (last) kp = prevkey(node, kp);
    PUSH(node, kp);
    KEYCPY((&ndx->key), kp);
	return ndx->key.seekadr;
}

SEEKADR xnext(INDEX *index)
{
	NODE *node;
    KEY *kp;
	
	ndx = index;
		
    /*  The state at return is that left by xfind:
          The top of the stack is the node in which a key was last found
          and the offset is that of the key.  */
		
    node = POP(kp);
    kp = NEXTKEY(kp);

    while (kp->ptr) {  /* Need to go down a level, at least */
        PUSH(node, kp);
        node = getnode(kp->ptr);
		kp = (KEY *)node->keys.keys;
        continue;
    }
	while (kp == (KEY *)(node->keys.keys + node->end_keys)) {
        if (STACKTOP == 0) return 0;
        node = POP(kp);
    }
    PUSH(node, kp);
    KEYCPY(&ndx->key, kp);
    return ndx->key.seekadr;
}

SEEKADR xprev(INDEX *index)
{
    NODE *node;
    KEY *kp;

    ndx = index;
    node = POP(kp);
    while (kp->ptr) {
        PUSH(node, kp);
        node = getnode(kp->ptr);
        kp = EXTRA(node);
        }
    kp = prevkey(node, kp);
    while (! kp) {
        if (STACKTOP) {
            node = POP(kp);
            kp = prevkey(node, kp);
        }
        else return 0;
        }
    PUSH(node, kp);
    KEYCPY(&ndx->key, kp);
    return kp->seekadr;
}

STATIC NODE * newnode(void)
{
	NODE *new = getnode(0L);   /* New slot in the cache */
    SEEKADR n;
	
	cache[0].dirty++;		   /* Make it pristine */
	if (n = ndx->hdr.freelist) {  /* If we can use an old node
                  -- that is, if the node freelist has an entry */
        xread(n, new, NODESIZE);
		ndx->hdr.freelist = new->seekadr;
        new->seekadr = n;
    }
    else {                        /* extend the file */
        new->seekadr = ndx->hdr.next;
        ndx->hdr.next += NODESIZE;
    }
    return new;
}

STATIC NODE * getnode(SEEKADR seekadr)
{
	int i;
    CACHE *c = cache, temp;
	
    for (i=0; i < NCACHE; i++, c++)     /* It may be in the cache */
		if (c->node->seekadr == seekadr && (c->handle == ndx->handle
			|| !seekadr)) break;
    if (i == NCACHE) {                  /* No luck */
        /* Try to find an empty slot */
        for (c=cache, i=0; i < NCACHE && c->node->seekadr ; c++, i++) ;
		/* None ... empty the last one */
        if (i == NCACHE) {  
            c--; i--;
            xwrite(c->node->seekadr, c->node, NODESIZE);
        }
        c->dirty = 0;
        if (! seekadr) memset(c->node, 0, NODESIZE);
          else xread(seekadr, c->node, NODESIZE);
    }
    temp = *c;                  /* Promote it to the top */
    temp.handle = ndx->handle;
	memmove(cache + 1, cache, i*sizeof(CACHE));
	*cache = temp;
	return temp.node;
}

STATIC void _push(NODE *node, KEY *kp)
{
    STACK *stk = &ndx->stack[STACKTOP++];
	
    if (STACKTOP > STACKDEPTH) fatal("Index stack overflow");
	stk->node = node->seekadr;
	stk->off = (char *)kp - (char *)node->keys.keys;
}

STATIC NODE * _pop(KEY **kp)
{
    STACK *stk = &ndx->stack[--STACKTOP];
	NODE *node = getnode(stk->node);
	
	*kp = (KEY *)(node->keys.keys + stk->off);
	return node;
}

STATIC int split(NODE * node)
{
    NODE *parent, *new;
    KEY *kp, *kpop, *old, *pivot;
    unsigned n, np;

	/* Find entry to be moved up a level */
	kp = (KEY *)(node->keys.keys + node->end_keys/2);
    pivot = prevkey(node, kp);
    np = KEYLEN(pivot);
    old = NEXTKEY(pivot);
	n = (char *)old - node->keys.keys;	/* The entry goes to new node, for */
    if (ndx->level) {				/*   the moment */
        /* If parent full, split it, then we will be back hopefully */
        parent = POP(kpop);
        if (KEYLEN(pivot) + parent->end_keys >= MAXDATA)
			return split(parent);
        }

    /* Copy first keys and pivot to new node */

	new = newnode();
	memcpy(new->keys.keys, node->keys.keys, n);
    new->end_keys = n - np;
	((KEY *)(new->keys.keys + n))->ptr = 0;

    /* Then move the remaining keys of the original node down */

    node->end_keys -= n;
	memmove(node->keys.keys, old, node->end_keys + sizeof(SEEKADR));
    cache[1].dirty = 1;

    /* If it was the root, create a new one */

    if (node->seekadr == ndx->hdr.root) {
        parent = newnode();
		kpop = (KEY *)parent->keys.keys;
        kpop->ptr = node->seekadr;
        ndx->hdr.root = parent->seekadr;
        }

    /* Put the pivot in the next higher level */

	pivot = (KEY *)((char *)new->keys.keys + new->end_keys);
    shiftup(parent, kpop, KEYLEN(pivot));
    KEYCPY(kpop, pivot);
    kpop->ptr = new->seekadr;   /* Pivot may have a ptr! */
    return cache[0].dirty = 1;
}

STATIC void shiftup(NODE *node, KEY *kp, unsigned n)
{
	int m = node->keys.keys + node->end_keys - (char *)kp + sizeof(SEEKADR);

    memmove((char *)kp + n, kp, m);
    node->end_keys += n;
	cache[0].dirty = 1;
}

STATIC void shiftdn(NODE *node, KEY *kp, unsigned n, unsigned leave)
{
	int m = node->keys.keys + node->end_keys - (char *)kp + sizeof(SEEKADR);
    SEEKADR link;

    if (! leave) memmove(kp, (char *)kp + n, m);
    node->end_keys -= n;
	cache[0].dirty = 1;
    if (node->end_keys || kp->ptr || ! STACKTOP) return;
    link = node->seekadr;
	node->seekadr = ndx->hdr.freelist;
    ndx->hdr.freelist = link;
    xwrite(link, node, NODESIZE);   /* Have to do this here! */
    cache[0].dirty = 0;
    POP(kp);
    kp->ptr = 0;
    STACKTOP++;
	cache[0].dirty = 1; 			/* POP changed cache[0] !!! */
}

STATIC initcache()
{
	int i;
	
	if (! cache_exists) {
		cache_exists++;
		for (i=0; i<NCACHE; i++)
			if (! (cache[i].node = calloc(1, NODESIZE+4))) return 0;
	}
	return 1;
}

STATIC xread(SEEKADR start, void *buffer, unsigned n)
{
    if (start != lseek(ndx->handle, start, 0L) ||
        n != read(ndx->handle, buffer, n))
        fatal("Read failure on file %s\n", ndx->filename);
}

STATIC xwrite(SEEKADR start, void *buffer, unsigned n)
{
    if (start != lseek(ndx->handle, start, 0L) ||
        n != write(ndx->handle, buffer, n))
        fatal("Write failure on file %s\n", ndx->filename);
}

STATIC KEY * prevkey(NODE *node, KEY *end)
{
	KEY *pk = (KEY *)node->keys.keys, *k = pk;

    if (pk == end) return NULL;
    while (k < end) {
        pk = k;
        k = NEXTKEY(k);
    }
    return pk;
}

STATIC void fatal(char *format, ...)
{
	va_list argptr;

	fprintf(stderr, "FATAL ERROR: ");
	va_start(argptr, format);
	vprintf(format, argptr);
	va_end(argptr);
	fputs("\n", stderr);
    exit(-1);
}

#ifdef TEST

STATIC void xprint(INDEX *index, char *filename)
{
	NODE *node = malloc(NODESIZE);
	unsigned long *blocks;
	int cblock = 0, last_block = 1;
    FILE *file = fopen(filename? filename: "CON", "w");

	blocks = calloc(20, sizeof(long));
    *blocks = index->hdr.root;
    if (! *blocks) {
        fputs("Empty index\n", file);
        return;
    }
    for (; cblock < last_block; cblock++) {
        KEY *kp, *end;

        xread(blocks[cblock], node, NODESIZE);
		kp = (KEY *)node->keys.keys;
		end = (KEY *)(node->keys.keys + node->end_keys);

        fprintf(file, "\nBlock at %lu\n", blocks[cblock]);
        for (; kp < end; (char *)kp += 2*sizeof(long) + strlen(kp->key)
            + 1) {
            if (kp->ptr)
                blocks[last_block++] = kp->ptr;
			fprintf(file, "         %-8lu %-8s %-8lu\n",
				kp->ptr, kp->key, kp->seekadr);
        }
        if (kp->ptr) {
            blocks[last_block++] = kp->ptr;
            fprintf(file, "         %-8lu\n", kp->ptr);
		}
    }
	free(node);
	free(blocks);
    fclose(file);
}

INDEX ix;
INDEX * index = &ix;
KEY ky;
KEY *entry = &ky;
int	seed = 16383;
unsigned random[10000];

main(int argc, char **argv)
{
    int n = atoi(argv[1]), i, j, m = (argc >= 2? atoi(argv[2]): 0);
    int fault = 0;

    if (argc < 2) {
		puts("\n\tUsage: NDX <number of keys> [<modulus>] [print]\n\n"
			"\tAny third argument causes the index to be printed\n"
			"\tUse of the second argument exercises the ISAM feature\n"
			"\n\tTry NDX 200 10 0, for instance\n");
		exit(-1);
		}
	srand(seed);
    for (i = 0; i < n; i++) random[i] = m? rand() % m: rand() + 1;
    index = xcreate("T.NDX", NULL, FIFO);
    for (i = 1; i <= n; i++) {
        sprintf(entry->key, "%u", random[i - 1]);
        entry->seekadr = i;
        xinsert(index, entry->key, entry->seekadr);
		}
    puts("Keys written");
    check(0, n, 0);
    check(0, n, 1);
    xclose(index);
    index = xopen("t.ndx", NULL);
    if (argc > 3) {xprint(index, NULL); xclose(index); return; }
    for (i = 1; i <= n; i++) {
        sprintf(entry->key, "%u", random[i - 1]);
		entry->seekadr = i;
        if (! xdelete(index, entry->key, entry->seekadr)) {
            printf("Delete failure: record %-u, data %-s\a\a\n", i, entry->key);
            output();
            fault++;
            break;
            }
          else if (xfindkeyrec(entry->key, entry->seekadr)) {
            printf("Key <%u, %u> not deleted!\n", i, random[i - 1]);
            output();
            fault++;
            break;
            }
    }
    xclose(index);
    if (! fault) puts("All keys successfully deleted");
}

check(int i, int j, int last)
{
	while (j--) {
		printf("%5u\r", j);
        if (! i++) xfirst(index, last);
        else if (! (last? xprev(index): xnext(index))) {
            printf("Key <%u, %u> not found\n", last? j + 1: i,
                    random[last? j: i - 1]);
            output();
            exit(-1);
            }
		}
    printf("All keys accessible on %s scan\n", last? "backward":
            "forward");
}

output()
{
    xprint(index, "t.tmp");
}

#endif


